function [f,opt_out] = create_signal(N,name,opt_arg)

  % Returns a signal of type specified by name.
  % The signal is complex and the second half of its Fourier transform is zero.
  % Its mean is also zero

  % Unless name is 'audio', N is the size of the created signal.

  if strcmp(name,'gaussian_noise')

    % gaussian random vector
    f = randn(N,1) + i*randn(N,1) ;

  elseif strcmp(name,'gaussian_noise_renorm')
    
    % gaussian random vector, renormalized in frequencies
    ff = [randn(N/2+1,1)+i*randn(N/2+1,1);zeros(N/2-1,1)] ;
    ff = ff./sqrt([1:N]') ;
    f = ifft(ff) ;
    
  elseif strcmp(name,'sinusoids')
    
    % sum of a few sinusoids
    ff = zeros(N,1) ;
    while (sum(abs(ff(2:end)))<1)
        ff(2:N/2+1) = (rand(N/2,1)< 1./[4:N/2+3]') ;
        f = N*ifft(ff.*exp(2*pi*i*rand(N,1)).*randn(N,1)) ;
    end

  elseif strcmp(name,'compact_sinusoids')
    
    % sum of a few sinusoids, multiplied by a window function
    ff = zeros(N,1) ;
    while (sum(abs(ff(2:end)))<1)
        ff(2:N/2+1) = (rand(N/2,1)< 1./[4:N/2+3]') ;
        f = N*ifft(ff.*exp(2*pi*i*rand(N,1)).*randn(N,1)) ;
    end
    win = window(N) ;
    f = f.*win(:) ;

  elseif strcmp(name,'piecewise_regular')

    % random line in an image
	 
    % name of the folder containing the images
    if (nargin>=3) && ~isempty(opt_arg)
       name_folder = opt_arg ;
    else
      % default
      name_folder = 'source_images' ;
    end

    % list the images contained in the directory source_images
    names_list = dir(name_folder) ;
    names_list = {names_list.('name')} ;
    model = ['image_'] ;
    nb_letters = numel(model) ;
    for s=numel(names_list):(-1):1
      if (numel(names_list{s})<nb_letters) || ...
         ~(strcmp(names_list{s}(1:nb_letters),model))
        names_list(s) = [] ;
      end
    end
    nb_candidates = numel(names_list) ;
    if (nb_candidates == 0)
      error(['create_signal : no images found.']) ;
    else
      % choose randomly an image from the list
      ind = 1+floor(nb_candidates*rand(1)) ;
      im = double(imread([name_folder,'/',names_list{ind}])) ;
      if (size(im,3)>1)
        im = mean(im,3) ;
      end
        
      % choose a position (ix,iy)
      ix = 1+floor(rand(1)*size(im,1)) ;
      iy = 1+floor(rand(1)*size(im,2)) ;
      % choose the slope of a line
      th = rand(1)*pi/2 ;
        
      % symmetrize the image, in order not to introduce additional
      % discontinuities
      im = [im,fliplr(im)] ;
      im = [im;flipud(im)] ;
        
      % list of indexes for the points on the line with slope th,
      % starting from (ix,iy)
      indx = ix+floor(sqrt(2)*sin(th)*[0:N-1]) ;
      indy = iy+floor(sqrt(2)*cos(th)*[0:N-1]) ;
      indx = mod(indx-1,size(im,1))+1 ;
      indy = mod(indy-1,size(im,2))+1 ;
        
      % extract indexes
      f = im(sub2ind(size(im),indx,indy)) ;
        
      % substract mean
      f = f-mean(f) ;
              
      % remove high frequencies
      ff = fft(f) ;
      ff(floor(3*N/8):N/2) = ff(floor(3*N/8):N/2) .* ...
			     cos((0:N/2-floor(3*N/8))*4/N*pi) ;
      f = ifft(ff) ;
      f = f(:) ;
    end

  elseif strcmp(name,'audio')

    % audio signal contained in a prescribed file
    % opt_arg is be the name of this prescribed file

    [f,opt_out] = wavread(opt_arg) ;
    f = f(:).' ;
    N = size(f,2) ;
    if (mod(N,2)==1)
      % N must be even
      N = N-1 ;
      f = f(1:N) ;
    end
    % remove high frequencies
    ff = fft(f) ;
    ff(floor(3*N/8):N/2) = ff(floor(3*N/8):N/2) .* ...
			   cos((0:N/2-floor(3*N/8))*4/N*pi) ;
    f = ifft(ff) ;
    f = f(:) ;
    
else

    error(['Function create_signal : the name of the function you ' ...
           'want is not recognized']) ;

end

% make sure that the second half of the Fourier transform is zero
ff = fft(f) ;
ff = [0;ff(2:N/2+1);zeros(N/2-1,1)] ;
f = ifft(ff) ;
% normalization
f = f / (1e-12 + max(abs(f))) ;

if (nargout>1) && ~exist('opt_out')
  % returns an empty output
  opt_out = [] ;
end
