function [freqs,psi,phi] = create_wavelets(N,a,name,p)

  % Inputs
  % N : size of the signals
  % a : dilation factor between wavelets
  %     (2 for dyadic wavelets)
  % name : name of the wavelet type
  %        ('simili-morlet' or 'cauchy' or 'mexican_hat'
  %         or 'gammatone')
  % p : optional parameter

  % Outputs
  % psi : a family of wavelets
  %       (as an array : each wavelet is a row)
  % phi : gaussian low frequency function
  % freqs : list of the characteristic frequencies of the wavelets

  % the frequencies are chosen so that the highest one is N/2 and the
  % lowest one is the largest possible below 1
  nbf = ceil(log(a*N/2)/log(a)) ;
  freqs = (N/2)*a.^(-[0:nbf-1]) ;

  ms = repmat(freqs.',[1,N]) ;
  oms = repmat([0:N-1],[nbf,1]) ;

  if strcmp(name,'simili-morlet')

    % Wavelets similar to morlet wavelets
    % Expression of the mother wavelet, in the Fourier domain:
    %      psi(om) = exp(-p (d.om-1)^2) - c. exp(-p (d.om)^2)
    % where c is chosen so that psi(0) = 0
    % and d is chosen so that the mother wavelet attains its maximum in 1

    if (nargin<4)
        p = 0.3 ;
    end

    % c such that psi(0)=0
    c = exp(-p) ;
    
    % we want the mother wavelet to attain its maximum in 1 so we need
    % to choose d so that psi'(1)=0
    % we thus need to find d such that d = 1 / (1-exp(-2pd))
    d = 1 ;
    for k=1:10
      d = 1/(1-exp(-p*2*d)) ;
    end
    
    psi = exp(-p*(d*oms./ms-1).^2)- c*exp(-p*(d*oms./ms).^2) ;
    
    max_psi = max(abs(psi(:))) ;
    psi = psi/max_psi ;

  elseif strcmp(name,'cauchy')

    % Cauchy wavelets with parameter p

    if (nargin<4)
      p = 3 ;
    end
    
    psi = (oms./ms).^p .* exp(p*(1-oms./ms)) ;

  elseif strcmp(name,'mexican_hat')

    % Mexican hat

    psi = exp(1) * (oms./ms).^2 .* exp(-(oms./ms).^2) ;

  elseif strcmp(name,'gammatone')

    % Gammatone wavelet, multiplied by om^n in the Fourier domain
    % The expression of the mother wavelet is:	 
    %     psi(om) = d * exp(c*(m-1)/b om i) * ...
    %               (c*om)^n / (b + i*(c*om-1))^m
    % where d and c are chosen so that |psi| is maximal in 1,
    % with maximal value 1

    if (nargin<4)
      b = 1 ;
      n = 1.5 ;
      m = 6 ;
    else
      b = p(1) ;
      n = p(2) ;
      m = p(3) ;
    end

    D = (m-2*n)^2 + 4*n*(b^2+1)*(m-n) ;
    c = (2*n-m - sqrt(D)) / (2 * (n-m)) ;

    d = (b + i*(c-1))^m / c^n ;

    psi = d * exp(c*(m-1)/b*oms./ms*i) .* (c*oms./ms).^n ...
	  ./ (b + i*(c*oms./ms-1)).^m ;
    
  else
    error(sprintf(['create_wavelets : the wavelets with name %s ', ...
		   'not implemented'],name)) ;
  end

  if (nargout>2)
    % low-frequency function
    phi = exp(-[0:N-1].^2) ;
  end
