function f_rec = ...
	 gradient_descent_final(fpsi_low,fpsi_high,products, ...
				psi_low,psi_high,opts)

  % This function is very similar to gradient_descent.
  % The main difference is that is optimizes all fpsi_high(s,:) and ...
  % fpsi_low(s,:) at the same time. Also, the parameter al is larger.
  % For more detailed comments, see gradient_descent.m

  nbf = size(fpsi_low,1) ; % number of frequencies
  N = size(fpsi_low,2) ; % size of the signal
  N2 = N/2+1 ;

  % normalization coefficients
  inv_max = repmat(1./max(abs(products),[],2),[1,N]) ;
  inv_max = min(inv_max,10*min(inv_max(:))) ;
  inv_max22N = inv_max.^2 * 2/N ;
        
  % functions used in the definition of g
  mult = [psi_high(1,:) ; abs(psi_high(2:nbf,:))+ ...
			  abs(psi_low(1:nbf-1,:)) ; psi_low(nbf,:)] ;
  mult_low = [psi_low(1:nbf-1,:)./(mult(2:nbf,:)+1e-15) ; ones(1,N)] ;
  mult_high = [ones(1,N) ; psi_high(2:nbf,:) ./ ...
			   (mult(2:nbf,:)+1e-15)] ;
    
  mult = mult(:,1:N2) ;
  mult_low = mult_low(:,1:N2) ;
  mult_high = mult_high(:,1:N2) ;
    
  fpsi_low_f = fft(fpsi_low,[],2) ;
  fpsi_high_f = fft(fpsi_high,[],2) ;

  % initial value for g    
  g = [fpsi_high_f(1,1:N) ; fpsi_high_f(2:nbf,1:N).* ...
			    exp(-i*angle(psi_high(2:nbf,:))) + ...
			    fpsi_low_f(1:nbf-1,1:N).* ...
			    exp(-i*angle(psi_low(1:nbf-1,:))) ; ...
					   fpsi_low_f(nbf,1:N)] ;
  g = g(:,1:N2) ;
    
  % initial value for the Fourier transform
  sum_mult = sum(abs(mult).^2,1) ;
  sum_mult(1) = 1 ;
  ff = sum(g.*conj(mult),1) ./ (sum_mult+1e-15) ;

  % second normalization coefficient    
  al =  2 * norm(products.*inv_max,'fro')^2 / norm(g(:),2)^2 ;

  % initialization of the loop
  step = 0 ;    
  k_it = 0 ;
  NotConverged = 1 ;
  time_cur = 0 ;
  time_init = toc ;
  % data structures used for l-bfgs
  m_bfgs = 3 ;
  last_g = zeros(size(g,1),size(g,2),m_bfgs) ;
  last_ff = zeros(m_bfgs,size(ff,2)) ;
  last_grads_g = zeros(size(g,1),size(g,2),m_bfgs) ;
  last_grads_ff = zeros(m_bfgs,size(ff,2)) ;
  nb_stored = 0 ;

  fprintf ('Gradient descent (final): ') ;

  % beginning of the loop
  while (k_it<opts.Kit_descent) && (time_cur<opts.time_limit) ...
	&& NotConverged
    
    % increment counter
    k_it = k_it+1 ;
        
    % print info
    if (mod(k_it,opts.nb_print_descent)==0)
      fprintf('%d - ',k_it) ;
    end

    % difference between true and current products
    fpsi_low_tmp = ifft(g(2:nbf+1,:).*mult_low,N,2) ;
    fpsi_high_tmp = ifft(g(1:nbf,:).*mult_high,N,2) ;
    diff_prod = fpsi_low_tmp.*conj(fpsi_high_tmp)-products ;
        
    % difference between the g(s,:) and the f star mult(s,:)
    diff_ff = bsxfun(@times,mult,ff) - g ;
       
    % current value of the objective
    obj(k_it) = norm(inv_max.*diff_prod,'fro')^2 + ...
		al * sum(sum(abs(diff_ff).^2))  ;

    % if the objective has increased, go back to the previous values
    % and take a smaller step
    if (k_it>1) && (obj(k_it)>=obj(k_it-1))
       
      obj(k_it) = obj(k_it-1) ;
      g = prev_g ;
      ff = prev_ff ;
            
      step = step/2 ;
            
    else
      
      % store g and ff
      prev_g = g ;
      prev_ff = ff ;
        
      % compute the gradient
      grad_g = zeros(nbf+1,N2) ;
      grad_tmp = fft(fpsi_high_tmp.*diff_prod ...
                     .* inv_max22N,[],2) ;
      grad_g(2:nbf+1,:) = grad_tmp(:,1:N2).*conj(mult_low) ;
      grad_tmp = fft(fpsi_low_tmp.*conj(diff_prod ...
                                    .* inv_max22N),[],2) ;
      grad_g(1:nbf,:) = grad_g(1:nbf,:) + ...
			grad_tmp(:,1:N2).*conj(mult_high) ;
      grad_g = grad_g - 2 * al * diff_ff ;

      grad_ff = 2 * al * sum(diff_ff.*conj(mult),1) ;

      % store the values
      if (norm(grad_g-last_grads_g(:,:,1),2)>1e-4 * norm(grad_g,2))
	last_g = circshift(last_g,[0,0,1]) ;
	last_g(:,:,1) = g ;
	last_ff = circshift(last_ff,[1,0]) ;
	last_ff(1,:) = ff ;
	last_grads_g = circshift(last_grads_g,[0,0,1]) ;
	last_grads_g(:,:,1) = grad_g ;
	last_grads_ff = circshift(last_grads_ff,[1,0]) ;
	last_grads_ff(1,:) = grad_ff ;
	nb_stored = nb_stored + 1 ;
      end
    end

    if (mod(k_it,1000)==1)
        
      % approximation of the hessian matrix

      % re-computation
      fpsi_low_tmp = ifft(g(2:nbf+1,:).*mult_low,N,2) ;
      fpsi_high_tmp = ifft(g(1:nbf,:).*mult_high,N,2) ;

      hess_g = zeros(nbf+1,N2) ;            
      h_tmp = diag(sum((abs(fpsi_high_tmp).*inv_max).^2,2)) * ...
              abs(mult_low).^2 * 2 / N^2 ;
      hess_g(2:(nbf+1),:) = h_tmp ;
      h_tmp = diag(sum((abs(fpsi_low_tmp).*inv_max).^2,2)) * ...
              abs(mult_high).^2 * 2 / N^2 ;
      hess_g(1:nbf,:) = hess_g(1:nbf,:) + h_tmp ;
      hess_g = 2*al + hess_g ;
            
      hess_ff = 2*al*sum(abs(mult).^2,1) ;
      hess_ff(1) = 1 ;

      h_inv_g = hess_g.^(-1) ;
      h_inv_ff = hess_ff.^(-1) ;
            
    end

    % l-bfgs: refine the approximation of the inverse hessian
    % copy the gradient
    grad_g2 = grad_g ;
    grad_ff2 = grad_ff ;
    if (nb_stored>=m_bfgs)

      last_g_diff = last_g(:,:,1:end-1) - last_g(:,:,2:end) ;
      last_ff_diff = last_ff(1:end-1,:) - last_ff(2:end,:) ;
      last_grads_g_diff = last_grads_g(:,:,1:end-1) - ...
			  last_grads_g(:,:,2:end) ;
      last_grads_ff_diff = last_grads_ff(1:end-1,:) - ...
			   last_grads_ff(2:end,:) ;

      r = real(sum(sum(conj(last_g_diff).*last_grads_g_diff,1),2)) ;
      r = r(:) + real(sum(conj(last_ff_diff).*last_grads_ff_diff,2)) ;

      % first loop
      for ii=1:m_bfgs-1
	if (r(ii)>1e-8)
	  alpha(ii) = (real(sum(sum(conj(last_g_diff(:,:,ii)) ...
				    .* grad_g2))) +  ...
		       real(sum(conj(last_ff_diff(ii,:)) ...
				.*grad_ff2))) / r(ii) ;
	else
	  alpha(ii) = 0 ;
	end
	grad_g2 = grad_g2 - alpha(ii) * last_grads_g_diff(:,:,ii) ;
	grad_ff2 = grad_ff2 - alpha(ii) * last_grads_ff_diff(ii,:) ;
      end

      % multiplication by the first approximation of the
      % inverse hessian
      grad_g2 = grad_g2.*h_inv_g ;
      grad_ff2 = grad_ff2.*h_inv_ff ;

      % second loop
      for ii=(m_bfgs-1):(-1):1
	if (r(ii)>1e-8)
	  be(ii) = (real(sum(sum(conj(last_grads_g_diff(:,:,ii)) ...
				 .* grad_g2))) +  ...
		  real(sum(conj(last_grads_ff_diff(ii,:)) ...
			   .*grad_ff2))) / r(ii) ;
	else
	  be(ii) = 0 ;
	end
	grad_g2 = grad_g2 + (alpha(ii)-be(ii)) * last_g_diff(:,:,ii) ;
	grad_ff2 = grad_ff2 + (alpha(ii)-be(ii)) * last_ff_diff(ii,:) ;
      end
    end

    if (mod(k_it,50)==1)
      step = 0.1 * min(norm(g(:),2)/norm(grad_g(:),2), ...
                       norm(ff(:),2)/norm(grad_ff(:),2)) ;
    end    
    
    % descent
    g = g - step * grad_g2 ;
    ff = ff - step * grad_ff2 ;
        
    % stopping criterium
    if (k_it>50) && (obj(k_it)>opts.stop_crit*obj(floor(k_it/2)))
      NotConverged = 0 ;
    end

    time_cur = toc - time_init ;
            
  end
  fprintf('\n%d iterations\n',k_it) ;

  f_rec = ifft(prev_ff,N) ;
