function f_rec = reconstruct(fpsi_noisy,psi,freqs_wav, ...
			     a,noise_descr,opts,debug)

  % Reconstructs a function from the modulus of its wavelet transform
  % The principle of the reconstruction is the following. For each
  % wavelet psi(j,:), we determine two wavelets psi_low(j,:) and
  % psi_high(j,:) such that the following product can be explicitely
  % computed from |f star psi(j,:)|:
  %     (f star psi_low(j,:)) x conj(f star psi_high(j,:))
  % Moreover, for each j, (f star psi_low(j,:)) can approximately be
  % estimated from (f star psi_high(j+1,:)).
  % Thus, after an initialization step, we can reconstruct all the
  % (f star psi_low(j,:)) and (f star psi_high(j,:)).

  % Inputs:
  % fpsi_noisy: modulus of the wavelet transform (possibly noisy)
  % psi: family of wavelets
  % freqs_wav: characteristic frequency of each wavelet
  % a: dilation factor
  % noise_descr:
  %   -> field ln: estimated amount of noise (in relative l2-norm)
  %   _> field noise_type: type of noise
  % opts: various parameters
  %   -> Kit_gs_rough: number of Gerchberg-Saxton iterations for a first
  %                    rough estimation of the signal
  %   -> Kit_gs: number of Gerchberg-Saxton iterations for the final
  %              refinement step
  %   -> Kit_descent: number of iterations in each gradient descent
  %   -> stop_crit: parameter for the stopping criterion in the gradient
  %                 descent (stops when the value of the objective
  %                 obj(kit) is larger than opts.stop_crit * obj(kit/2))
  %   -> nb_print_descent: during a gradient descent, a message will be
  %                        printed every nb_print_descent iteration
  %   -> nb_print_gs: during the Gerchberg-Saxton algorithm, a message will
  %      be printed every nb_print_gs iteration
  %   -> time_limit: maximal amount of time spent for each gradient descent
  %   -> reestimate: whether to execute the reestimation procedure or not
  %   -> exp_decay: in practice, the algorithm works only if the wavelets
  %                 decrease at exponential speed (in the Fourier domain);
  %                 in practice, it can approximately work for non exponen-
  %                 tially decreasing wavelets but it requires an appro-
  %                 priate truncation step; exp_decay = 'yes' indicates
  %                 that the wavelets decrease exponentially (and thus, no
  %                 truncation will be performed), exp_decay = 'no'
  %                 indicates that they do not decrease fast enough (a
  %                 truncation will be performed), with exp_decay = 'auto',
  %                 the program determines by itself whether they decrease
  %                 exponentially or not
  %   -> only_gs: if true, reconstructs the signal from low to high
  %               frequencies by using Gerchberg-Saxton only at each step
  % debug: debugging options (optional)
  %   -> bool: whether to perform debugging or not
  %   -> function: name of the function which must be debugged
  %                ('reconstruct' or 'reestimate')
  %   -> num_params: parameters (array of numbers)
  %                  (in the case where debug.function == 'reconstruct',
  %                   num_params(1) is the scale at which to debug ; at
  %                   this scale, the debugging will be performed before
  %                   the first gradient descent if num_params(2) is 1,
  %                   after the first gradient descent if num_params(2) is
  %                   2 and after the second gradient descent if
  %                   num_params(2) is 3)
  %   -> true_f: true signal

  % Output:
  % f_rec: reconstructed function

  if (nargin<7)
     debug.bool = false ;
  end

  nbf = numel(freqs_wav) ; % number of frequencies
  N = size(psi,2) ; % size of the signal

  if (nbf<3)
     error(['Function reconstruct: the reconstruction process is ', ...
	    'not implemented for less than three wavelets in the family']) ;
  end

  % set unprovided option parameters to default values
  opts = fill_opts(opts,{'Kit_gs_rough','Kit_gs','Kit_descent', ...
			 'stop_crit','nb_print_descent','nb_print_gs', ...
			 'time_limit','reestimate','exp_decay', ...
			 'only_gs'}, ...
		   {10,2000,3000,0.995,1000,500,Inf,true,'auto',false}) ;
  % check whether the wavelets decay exponentially
  exp_decay = exp_decay_approx(psi,a,opts.exp_decay) ;

  % the expression of psi_low(j,:) and psi_high(j,:) will be:
  %    psi_low(j,om) = psi(j,om) * exp(-alpha*om*a^(j-1))
  %    psi_high(j,om) = psi(j,om) * exp(alpha*om*a^(j-1))

  % compute a very rough approximation of the signal, to be used in
  % the estimation of some parameters
  f_ref = gerchberg_saxton ...
            ([],fpsi_noisy,psi,opts.Kit_gs_rough,opts.nb_print_gs) ;
  f_ref_f = fft(f_ref) ;
  f_ref_f(1) = 0 ;
  f_ref_f(N/2+2:end) = 0 ;
  f_ref = ifft(f_ref_f) ;

  % wavelet transform of f_ref
  fpsi_ref = wavelet_transform(f_ref,psi) ;
  % add noise
  fpsi_noisy_ref = add_global_noise(noise_descr.ln, ...
				    noise_descr.noise_type, ...
				    abs(fpsi_ref),[],psi) ;

  % reference relative l2-errors induced by the noise over the ...
  % squared moduli
  err_ref = sqrt(sum(abs(abs(fpsi_noisy_ref).^2-abs(fpsi_ref).^2).^2,2)) ./ ...
	   sqrt(sum(abs(fpsi_ref).^4,2)) ;
  inds_ref = find(err_ref<1e-3) ;
 
  if ~opts.only_gs
    % choose alpha

    % alpha is computed on a middle-frequency wavelet (with index j0),
    % to avoid problems due to truncation in frequency (for the
    % high-frequency wavelets) or aliasing (for the low-frequency
    % wavelets)
    j0 = min(nbf-2,round(sqrt(nbf))) ;
    alpha = find_alpha(psi(j0,:),psi(j0+1,:),psi(j0+2,:), ...
		       a,exp_decay,j0) * a^(-(j0-1)) ;

    % this value has been chosen independently of the noise
    % for small noise levels, taking a larger value can be better
    if ~isempty(inds_ref)
      fprintf('Choosing the best alpha\n') ;
      alphas = alpha * 2.^[0:0.1:1] ; % possible values for alpha
      for n_alpha = 1:numel(alphas)
	      
	% for each value of alpha, estimate (on f_ref) the error which
	% the presence of noise will induce in the computation of the
	% products (f star psi_low(s,:)) * conj(f star psi_high(s,:))
	  
	% we will select the highest value of alpha for which this
	% error is not too high
	  
	% compute the corresponding psi_low and psi_high
	[psi_low,psi_high] = compute_low_and_high(psi,alphas(n_alpha), ...
			     a,exp_decay) ;
	% renormalization
	cs = sqrt(max(abs(psi_high),[],2)./max(abs(psi_low),[],2)) ;
	psi_low = bsxfun(@times,psi_low,cs) ;
	psi_high = bsxfun(@times,psi_high,cs.^(-1)) ;
      
	% wavelet transform of f_ref
	fpsi_ref = wavelet_transform(f_ref,psi) ;
	fpsi_ref_low = wavelet_transform(f_ref,psi_low) ;
	fpsi_ref_high = wavelet_transform(f_ref,psi_high) ;
	% add noise
	fpsi_noisy_ref = add_global_noise(noise_descr.ln, ...
					  noise_descr.noise_type, ...
					  abs(fpsi_ref),[],psi) ;
      
	% compute the errors on the products
	for s = inds_ref'
	  true_prod_ref(s,:) = fpsi_ref_low(s,:) ...
			       .*conj(fpsi_ref_high(s,:)) ;
	  % choice of the parameter
	  [forg,err_prod] = ...
	  eval_prod_ref(abs(fpsi_noisy_ref(s,:)).^2, ...
			alphas(n_alpha) * a^(s-1), ...
			true_prod_ref(s,:)) ;
	  % error (in relative l2-norm)
	  err_tmp(s) = err_prod / norm(true_prod_ref(s,:),2) ;
	end

      % error associated to this value of alpha
      err_alpha(n_alpha) = max(err_tmp) ;

    end

    % index of the largest alpha for which the error is not larger
    % that 1 percent
    [forg,n_alpha] = min([err_alpha<1e-2,0]) ;
    n_alpha = max(1,n_alpha-1) ;
    % print info
    fprintf('alpha has been multiplied by %.2f\n', ...
	    alphas(n_alpha)/alpha) ;
    % update the value of alpha
    alpha = alphas(n_alpha) ;
    end

  else
    % no intermediate wavelets in the case where the reconstruction is
    % performed with Gerchberg-Saxton only
    alpha = 0 ;
  end

  % compute psi_low and psi_high
  [psi_low,psi_high] = compute_low_and_high(psi,alpha,a,exp_decay) ;

  % renormalization
  cs = sqrt(max(abs(psi_high),[],2)./max(abs(psi_low),[],2)) ;
  psi_low = bsxfun(@times,psi_low,cs) ;
  psi_high = bsxfun(@times,psi_high,cs.^(-1)) ;
  
  % wavelet transform of the very rough approximation
  fpsi_ref = wavelet_transform(f_ref,psi) ;
  fpsi_ref_low = wavelet_transform(f_ref,psi_low) ;
  fpsi_ref_high = wavelet_transform(f_ref,psi_high) ;
  % add noise
  fpsi_noisy_ref = add_global_noise(noise_descr.ln, ...
				    noise_descr.noise_type, ...
				    abs(fpsi_ref),[],psi) ;

  % compute (f star psi_low(j,:)) x conj(f star psi_high(j,:)) for
  % each j
  for s=1:nbf
    % the computation is performed by the function eval_prod. It needs
    % a regularization parameter to be provided. The value of this
    % parameter is determined by testing several values on the rough
    % approximated signal and taking the one that performs best.      
    true_prod_ref(s,:) = fpsi_ref_low(s,:) ...
			 .*conj(fpsi_ref_high(s,:)) ;
    % choice of the parameter
    ind = eval_prod_ref(abs(fpsi_noisy_ref(s,:)).^2, ...
			alpha * a^(s-1), ...
			true_prod_ref(s,:)) ;
    % estimation of the product
    products(s,:) = eval_prod(abs(fpsi_noisy(s,:)).^2, ...
                              alpha * a^(s-1),ind) ;
  end   
    
  % beginning of the iterative reconstruction of f star psi_low(j,:)
  % and f star psi_high(j,:)
  
  % initialization
  fpsi_low = zeros(nbf,N) ;
  fpsi_high = zeros(nbf,N) ;
    
  for s=1:nbf
    % size of the approximate support of psi_low(s,:) (in Fourier)
    supp_psi_low(s) = estimate_ind(psi_low(s,:)) ;
    supp_psi_high(s) = estimate_ind(psi_high(s,:)) ;
  end
    
  % reconstruct f star psi_low(nbf,:) (lowest frequency)
  fpsi_low(nbf,:) = eval_lowest(fpsi_noisy(nbf-1,:), ...
                                fpsi_noisy(nbf,:), ...
                                psi(nbf-1,:),psi(nbf,:), ...
                                psi_low(nbf,:)) ;

  % deduce f star psi_high(nbf,:)
  fpsi_high(nbf,:) = conj(products(nbf,:)./fpsi_low(nbf,:)) ;
    
  fprintf('Scale by scale estimation\n') ;
  for S=(nbf-1):(-1):1
  
    % from f star psi_high(S+1), compute f star psi_low(S,:)
    fpsi_low(S,:) = estimate(S,psi_low,psi_high,fpsi_low, ...
			     fpsi_high,supp_psi_high(S+1)) ;

    if opts.only_gs
      psi_conj = dual_wavelets(psi(S:end,1:supp_psi_low(S))) ;
      psi_conj = [psi_conj,zeros(nbf-S+1,N-supp_psi_low(S))] ;
    end

    % debugging
    if debug.bool && strcmp(debug.function,'reconstruct') ...
       && (S == debug.num_params(1)) && (1 == debug.num_params(2))
      break ;
    end


    if (S>=2)
      tic ;
      if ~opts.only_gs
	% adjust the values
	[fpsi_low,fpsi_high] = ...
	gradient_descent(fpsi_low,fpsi_high,products, ...
			 psi_low,psi_high,S, ...
			 supp_psi_high(S+1), ...
			 opts) ;
	% debugging
	if debug.bool && strcmp(debug.function,'reconstruct') ...
	   && (S == debug.num_params(1)) && (2 == debug.num_params(2))
	  break ;
	end
      else
	% case where all must be done with Gerchberg-Saxton
	fprintf('Scale %d\n',S) ;
	% from the current values of fpsi_low, fpsi_high, deduce an
	% estimation of the signal f
	f_rec = ifft(sum(fft(fpsi_low(S:end,:),[],2).*psi_conj,1)) ;
	% optimize with Gerchberg-Saxton
	f_rec = gerchberg_saxton(f_rec,fpsi_noisy,psi, ...
				 opts.Kit_gs,opts.nb_print_gs, ...
				 supp_psi_low(S)) ;
	% deduce fpsi_low, fpsi_high
	fpsi_low(S:end,:) = wavelet_transform(f_rec,psi(S:end,:)) ;
	fpsi_high(S+1:end,:) = wavelet_transform(f_rec,psi(S+1:end,:)) ;
      end
      toc ;
      if opts.reestimate
	% try to correct the errors
	[fpsi_low,fpsi_high] = reestimate ...
				 (fpsi_low,fpsi_high,psi, ...
				  psi_low,psi_high,fpsi_noisy, ...
				  S,products, ...
				  supp_psi_low,supp_psi_high, ...
				  debug) ;
      end
    end
        
    % improve the values found until now by gradient descent
    tic ;
    if ~opts.only_gs
      [fpsi_low,fpsi_high] = ...
      gradient_descent(fpsi_low,fpsi_high,products, ...
		       psi_low,psi_high,S, ...
		       supp_psi_high(S+1), ...
		       opts) ;
    else
      % Gerchberg-Saxton (as previously)
      fprintf('Scale %d\n',S) ;
      fpsi_low(S+1,:) = fpsi_high(S+1,:) ;
      f_rec = ifft(sum(fft(fpsi_low(S:end,:),[],2).*psi_conj,1)) ;
      f_rec = gerchberg_saxton(f_rec,fpsi_noisy,psi, ...
			       opts.Kit_gs,opts.nb_print_gs, ...
			       supp_psi_low(S)) ;
      fpsi_low(S:end,:) = wavelet_transform(f_rec,psi(S:end,:)) ;
      fpsi_high(S+1:end,:) = wavelet_transform(f_rec,psi(S+1:end,:)) ;	
    end
    toc ;
            
    % compute f star psi_high(S,:)
    fpsi_high(S,:) = conj(products(S,:)./fpsi_low(S,:)) ;
        
    if (S>1)
      % f star psi_high(S,:) should be negligible at frequencies
      % greater than sup_psi_high(S)
      % set these negligible values to zero (to avoid error propagation)
      fpsi_high_f = fft(fpsi_high(S,:)) ;
      fpsi_high(S,:) = ifft(fpsi_high_f(1:supp_psi_high(S)),N) ;
    end

    % debugging
    if debug.bool && strcmp(debug.function,'reconstruct') ...
       && (S == debug.num_params(1)) && (3 == debug.num_params(2))
       break ;
    end
        
  end    
        
  if debug.bool && strcmp(debug.function,'reconstruct')
    % debugging
    plot_debug_info(debug.true_f,fpsi_low,fpsi_high,products, ...
		    psi_low,psi_high)
  end
    
  if ~opts.only_gs
    % final gradient descent
    tic ;
    f_rec = gradient_descent_final ...
	      (fpsi_low,fpsi_high,products,psi_low, ...
	       psi_high,opts) ;
    toc ;
  end
  
  % improve the reconstruction with Gerchberg-Saxton
  f_rec = gerchberg_saxton(f_rec,fpsi_noisy,psi, ...
			   opts.Kit_gs,opts.nb_print_gs) ;


function opts = fill_opts(opts,field_names,values)

  % for each name in 'field_names', checks is opts has a field with
  % this name ; if not, creates such a field and fills it with the
  % value given in 'values'

  for k=1:numel(field_names)
      if ~(isfield(opts,field_names{k}))
	 opts.(field_names{k}) = values{k} ;
      end
  end

function plot_debug_info(f,fpsi_low,fpsi_high,products, ...
			  psi_low,psi_high)
  % plot informations for debugging
     
  % compute the true f star psi_low and f star psi_high
  ff = fft(f) ;
  fpsi_low_true = wavelet_transform(f,psi_low) ;
  fpsi_high_true = wavelet_transform(f,psi_high) ;
  
  figure (1) ;
  clf ; hold on ;
  draw_functions(abs(fft(fpsi_low_true,[],2)),'r') ;
  draw_functions(abs(fft(fpsi_low,[],2)),'b') ;
  ax = findobj(gcf,'type','axes');
  pos = cell2mat(get(ax,'position'));
  ncols = numel(unique(pos(:,1)));
  nrows = numel(unique(pos(:,2)));
  subplot(nrows,ncols,ceil(ncols/2)) ;
  title(['Difference between true (red) and reconstructed ', ...
	 '(blue) fpsi low in Fourier']) ;

  figure (2) ;
  clf ; hold on ;
  draw_functions(abs(fpsi_low_true),'r') ;
  draw_functions(abs(fpsi_low),'b') ;
  subplot(nrows,ncols,ceil(ncols/2)) ;
  title(['Difference between true (red) and reconstructed ', ...
	 '(blue) fpsi low in time']) ;

  figure (3) ;
  clf ; hold on ;
  draw_functions(angle(fpsi_low./fpsi_low_true),'b') ;
  subplot(nrows,ncols,ceil(ncols/2)) ;
  title(['Difference of phase between true and ', ...
	 'reconstructed fpsi low']) ;

  figure (4) ;
  clf ; hold on ;
  draw_functions(abs(fft(fpsi_high_true,[],2)),'r') ;
  draw_functions(abs(fft(fpsi_high,[],2)),'b') ;
  subplot(nrows,ncols,ceil(ncols/2)) ;
  title(['Difference between true (red) and reconstructed ', ...
	 '(blue) fpsi high in Fourier']) ;

  figure (5) ;
  clf ; hold on ;
  draw_functions(abs(fpsi_high_true),'r') ;
  draw_functions(abs(fpsi_high),'b') ;
  subplot(nrows,ncols,ceil(ncols/2)) ;
  title(['Difference between true (red) and reconstructed ', ...
	 '(blue) fpsi high in time']) ;

  % compute true products
  true_prods = fpsi_low_true.*conj(fpsi_high_true) ;

  figure (6) ; clf ; hold on ;
  draw_functions(real(fpsi_low.*conj(fpsi_high)),'g') ;
  draw_functions(real(true_prods),'r') ;
  draw_functions(real(products),'b') ;
  subplot(nrows,ncols,ceil(ncols/2)) ;
  title(['Difference between true (red), computed (blue) and ', ...
	 'reconstructed (green) products']) ;

  error('reconstruct.m : stopped for debugging') ;
