function [fpsi_low,fpsi_high] = reestimate(fpsi_low,fpsi_high,psi, ...
					   psi_low,psi_high,fpsi_noisy, ...
					   S,products, ...
					   supp_psi_low,supp_psi_high, ...
					   debug)

  % This function is called by reconstruct.m just after the values of
  % f star psi_high(S-1,:) and f star psi_low(S,:) have been
  % estimated. It tries to detect and correct the errors in these
  % estimations.

  % Inputs
  % fpsi_low: values of f star psi_low estimated after the last
  %           reconstruction step
  % fpsi_high: idem for f star psi_high
  % psi: wavelets
  % psi_low, psi_high: intermediate wavelets used in reconstruct.m
  % fpsi_noisy: phaseless measurements which reconstruct.m has received
  %             as input
  % S: index of the components to reestimate
  % products: for any s, (f star psi_low(s,:)) * conj(f star psi_high(s,:))
  % supp_psi_low,supp_psi_high: approximate sizes of the support of
  %                             the psi_low's and psi_high's (in Fourier)

  % Ouputs
  % fpsi_low: corrected f star psi_low's
  % fpsi_high: corrected f star psi_high's
    
  % Principle: the signals are divided in (overlapping) small windows.
  % The reestimation is done window by window, by an exhaustive-search
  % method.

  N = size(fpsi_noisy,2) ; % size of the signal
  nbf = size(fpsi_noisy,1) ; % number of frequencies

  if debug.bool && strcmp(debug.function,'reestimate') ...
     && (S == debug.num_params(1))
    % debugging
    % store the initial value of f star psi_low(S,:), so that in the
    % end we can compare this value with the reestimated value
    fpsi_low_init = fpsi_low(S,:) ;
  end


  % determine the size of the windows
  for s=S-1:S
    max_psi = max(abs(psi(s,:))) ;
    ind = max([1:N].*(abs(psi(s,:))>0.02*max_psi)) ;
    win_size(s) = min(N,floor(13*N/ind)) ;
    win_size(s) = 2*round(win_size(s)/2) ; % the size must be even
  end


  % determine where there are errors which must be corrected

  % modulus of f star psi(S+1,:)
  fpsi_mod = abs(fpsi_noisy(S+1,:)) ;

  somme = abs(psi_high(S+1,:)).^2 + abs(psi_low(S+1,:)).^2 ;
  somme(1) = 1 ;
  mul1 = psi(S+1,:).*conj(psi_low(S+1,:)) ./ (somme+1e-12) ;
  mul2 = psi(S+1,:).*conj(psi_high(S+1,:)) ./ (somme+1e-12) ;
  % value of f star psi(S+1,:) given by the current estimations of
  % f star psi_high(S+1,:) and f star psi_low(S+1,:)
  fpsi_est = ifft(fft(fpsi_low(S+1,:)).*mul1 + ...
		  fft(fpsi_high(S+1,:)).*mul2) ;
  % modulus of this estimated f star psi(S+1,:)
  fpsi_mod_est = abs(fpsi_est) ;

  % fpsi_mod_est and fpsi_mod must be approximately equal. The
  % reestimation will be performed at the points where they are not
  % approximately equal

  % points where fpsi_mod and fpsi_mod_est are approximately equal
  bools = (abs(fpsi_mod-fpsi_mod_est)<0.5*abs(fpsi_mod)) ;
  % computes bools2: bools2(k) indicates how many points in a
  % neighborhood of k of size win_size(S) satisfy the approximate
  % equality
  supp = zeros(1,N) ; supp(1:win_size(S)/2) = 1 ;
  supp(end-win_size(S)/2+1:end) = 1 ;
  bools2 = ifft(fft(bools).*fft(supp)) ;
  % we reestimate at points where at least 10 percent of the neighbors
  % do not satisfy the approximate equality
  bools = (abs(bools2)<sum(supp)*0.9) ;
  % add to these points their neighbors in a neighborhood of the size
  % of the window
  bools2 = ifft(fft(bools).*fft(supp)) ;
  bools = (abs(bools2)>1/2) ;
 
  inds_rees = find(bools) ;
  
  % reestimate
  if (numel(inds_rees)>0)

    if (win_size(S-1)>=N)
      % First case: the frequency is sufficiently low so that the
      % division in windows is not necessary.
       
      fprintf('Reinitialization at scale %d\n',S) ;

      % exhaustive-search method
      fpsi_bis = eval_lowest(fpsi_noisy(S,:), ...
                             fpsi_noisy(S+1,:), ...
                             psi(S,:),psi(S+1,:), ...
                             [psi_low(S,:);psi_high(S+1,:)]) ;
      % multiply by a global phase
      fpsi_low(S,:) = fpsi_bis(1,:) * ...
		      exp(i*angle(sum(conj(fpsi_bis(1,:)).*fpsi_low(S,:)))) ;
      fpsi_high(S+1,:) = fpsi_bis(2,:) * ...
			 exp(i*angle(sum(conj(fpsi_bis(2,:)) ...
					 .*fpsi_high(S+1,:)))) ;
      
      if debug.bool && strcmp(debug.function,'reestimate') ...
	 && (S == debug.num_params(1))
        % debugging
        fpsi_true = wavelet_transform(debug.true_f,psi_high(S+1,:)) ;
        figure (1) ; clf ; hold on ;
        plot(abs(fpsi_true)) ;
        plot(abs(fpsi_high(S+1,:)),'r') ;
	title('true (blue) and estimated (red) fpsi high') ;
	error('reestimate.m: stopped for debugging') ;
      end
    
    else
      % Second case: the division in windows is necessary
    
      % Principle:
      % Around each point s at which a reestimation is necessary, we
      % consider f * win(.-s), where f is the unknown signal and win
      % is the window (centered around zero).
      % If win is sufficiently smooth, then, for s = S,S-1 :
      %     (f * win(.-s)) star psi(s,:)  \approx
      %          (f star psi(s,:)) * win(.-s)
      % So from the modulus of f star psi(s,:), we can approximately
      % compute the modulus of (f * win(.-s)) star psi(s,:). Thus, we
      % can express f * win(.-s) as the solution of a phase retrieval
      % problem. If the size of win is not too large, then this phase
      % retrieval problem can be solved by an exhaustive search
      % method. Solving it gives an estimation of (f star psi(s,:)) *
      % win(.-s).

      Nw = win_size(S) ; % shortcut for the size of the window
      
      psi_time = ifft(psi,N,2) ; % wavelets in the time domain

      % wavelets truncated to the size of the window
      psi1 = [psi_time(S,1:Nw/2) , psi_time(S,N-Nw/2+1:N)] ;
      psi2 = [psi_time(S-1,1:Nw/2) , psi_time(S-1,N-Nw/2+1:N)] ;

      psi2_f = fft(psi2) ;
      psi1_f = fft(psi1) ;
      
      % window
      win = window(Nw) ;
   
      % The approximate equality
      %     (f * win(.-s)) star psi(S,:)  \approx
      %          (f star psi(S,:)) * win(.-s)
      % is not very precise (win is not smooth enough). Instead, we
      % use a refined approximation formula for the modulus of (f *
      % win(.-s)) star psi(s,:). This formula involves several
      % parameters which are computed by the function
      % modulus_app_parameters
      [shifts1,shifts2,shifts1_inv,mul1,mul2,mul1_inv] = ...
      modulus_app_parameters(win,psi1,psi2) ;
    
      % decompose the indexes at which a reestimation must be
      % performed in disjoint intervals ("blocks")
      inds_beg = find(inds_rees~=[-1,inds_rees(1:end-1)]+1) ;
      inds_end = find(inds_rees~=[inds_rees(2:end),N+2]-1) ;
      inds_beg = inds_rees(inds_beg) ; % beginnings of blocks
      inds_end = inds_rees(inds_end) ; % ends of blocks
      % merge first and last blocks (the problem is circular)
      if (numel(inds_beg)>1) && (inds_end(end) == N) && (inds_beg(1) == 1)
	inds_beg(1) = inds_beg(end)-N ;
	inds_beg = inds_beg(1:end-1) ;
	inds_end = inds_end(1:end-1) ;
      end

      nb_blocks = numel(inds_beg) ;
      % space between two consecutive positions of the window
      spa = max(1,floor(Nw/3)) ;

      % info
      fprintf('Reinitialization at scale %d : %d blocks\n', ...
              S,nb_blocks) ;
    
      not_break = 1 ;

      % perform the reestimation over each block    
      for s=1:nb_blocks

	% adjusts the space between consecutive positions of the
	% window
	spa_s = (inds_end(s)-inds_beg(s))/ ...
		ceil((inds_end(s)-inds_beg(s))/spa) ;
	% list of the positions of the window
	inds_c = round(inds_beg(s):spa_s:inds_end(s)) ;

	% info
	fprintf('  Block %d : %d subblocks ',s,numel(inds_c)) ;

	% for each position, reestimate
	for s2=1:numel(inds_c)
	  % info
          tic_toc_print('(%d/%d) ',s2,numel(inds_c)) ;
	  
	  % support of the window          
          i1 = inds_c(s2)-round(Nw/2) ;
          i2 = i1+Nw-1 ;
          ind_trunc = mod((i1:i2)-1,N)+1 ;

	  % approximation of the modulus of (f * win) star psi1 and (f
	  % * win) star psi2
          fpsi1_app = fpsi_noisy(S,mod(ind_trunc+ ...
				       round(shifts1)-1,N)+1).*abs(mul1) ;
          fpsi2_app = fpsi_noisy(S-1,mod(ind_trunc+ ...
					 round(shifts2)-1,N)+1).*abs(mul2) ;
          
	  if debug.bool && strcmp(debug.function,'reestimate') ...
	     && (S == debug.num_params(1)) ...
	     && (s == debug.num_params(2)) ...
	     && (s2 == debug.num_params(3))
	    % debugging
	    
	    fprintf('\n center of the window: %d\n',inds_c(s2)) ;
	    % compute true (f * win) star psi1 and
	    % (f * win) star psi2
            f_trunc = debug.true_f(ind_trunc).'.*win ;
            fpsi1_true = wavelet_transform(f_trunc,fft(psi1)) ;
            fpsi2_true = wavelet_transform(f_trunc,fft(psi2)) ;
            figure (1) ; clf ;
            subplot(121) ; hold on ;
            plot(abs(fpsi1_app),'r') ;
            plot(abs(fpsi1_true)) ;
	    title('true (blue) and estimated (red) abs(fw star psi1)') ;
            subplot(122) ; hold on ;
            plot(abs(fpsi2_app),'r') ;
            plot(abs(fpsi2_true)) ;
	    title('true (blue) and estimated (red) abs(fw star psi2)') ;
          end
          
	  % computes all possible (f*win) star psi1 and
	  % (f*win) star psi2
          sol1 = solutions(fpsi1_app,fft(psi1),14) ;
	  sol2 = solutions(fpsi2_app,fft(psi2),24) ;
            
	  if debug.bool && strcmp(debug.function,'reestimate') ...
	     && (S == debug.num_params(1)) ...
	     && (s == debug.num_params(2)) ...
	     && (s2 == debug.num_params(3))	     
	    % debugging

	    % find the line in sol1 which is the closest (in l2 norm)
	    % to the true value
            fpsi1_true = fpsi1_true * exp(-i*angle ...
                                               (fpsi1_true(Nw/2))) ;
            fpsi1_rec_tmp = ifft(sol1,Nw,2) ;
            fpsi1_rec_tmp = bsxfun(@times,fpsi1_rec_tmp, ...
                               exp(-i*angle(fpsi1_rec_tmp(:,Nw/2))));
            dist1 = bsxfun(@minus,fpsi1_rec_tmp,fpsi1_true) ;
            [forg,ind1] = min(sum(abs(dist1).^2,2)) ;
	    fpsi1_rec_tmp = fpsi1_rec_tmp(ind1,:) ;
	    
	    % same for sol2
            fpsi2_true = fpsi2_true * exp(-i*angle ...
					       (fpsi2_true(Nw/2))) ;
	    rand_ind = randsample(size(sol2,1), ...
				  min(10000,size(sol2,1))) ;
            fpsi2_rec_tmp = ifft(sol2(rand_ind,:),Nw,2) ;
            fpsi2_rec_tmp = bsxfun(@times,fpsi2_rec_tmp, ...
			       exp(-i*angle(fpsi2_rec_tmp(:,Nw/2))));
            dist2 = bsxfun(@minus,fpsi2_rec_tmp,fpsi2_true) ;
            [forg,ind2] = min(sum(abs(dist2).^2,2)) ;
	    fpsi2_rec_tmp = fpsi2_rec_tmp(ind2,:) ;
	  
            figure (2) ; clf ;
            subplot(321) ; hold on ;
            plot(abs((fpsi1_true))) ;
            plot(abs((fpsi1_app)),'g') ;
            plot(abs((fpsi1_rec_tmp)),'r') ;
	    title('Best solution found for psi1 (modulus)') ;
            subplot(323) ; hold on ;
            plot(abs(fft(fpsi1_true))) ;
            plot(abs(fft(fpsi1_rec_tmp)),'r') ;
	    title('Best solution found for psi1 (Fourier)') ;
            subplot(325) ; hold on ;
	    plot(unwrap(angle(fpsi1_true./fpsi1_rec_tmp))) ;
	    title('Best solution found for psi1 (phase diff)') ;
            subplot(322) ; hold on ;
            plot(abs((fpsi2_true))) ;
            plot(abs((fpsi2_app)),'g') ;
            plot(abs((fpsi2_rec_tmp)),'r') ;
	    title('Best solution found for psi2') ;
            subplot(324) ; hold on ;
            plot(abs(fft(fpsi2_true))) ;
            plot(abs(fft(fpsi2_rec_tmp)),'r') ;
	    title('Best solution found for psi2 (Fourier)') ;
            subplot(326) ; hold on ;
	    plot(unwrap(angle(fpsi2_true./fpsi2_rec_tmp))) ;
	    title('Best solution found for psi2 (phase diff)') ;
                
          end
            
	  % From the list of all possible (f*win) star psi1 and
	  % (f*win) star psi2, tries to determine the real solutions

	  % There are too many possible solutions for (f*win) star
	  % psi2. To increase the speed of the code, we apply a simple
	  % test to eliminate a large part of the solutions.

	  % Each line of sol2 represents a possible choice for (f*win)
	  % star psi2. We deconvolve by psi2 and convolve by psi1 to
	  % get the corresponding (f*win) star psi1.
	  mult = psi1_f(1:size(sol2,2)) ./ ...
		 (psi2_f(1:size(sol2,2))+1e-13) ;
	  fpsi1_from_sol2 = bsxfun(@times,sol2,mult) ;
	  nb = min(60,Nw) ; % subsampling
	  fpsi1_from_sol2 = ifft(fpsi1_from_sol2,nb,2) ;
	  % mod_comp is the subsampled modulus of (f*win) star psi1
	  % (estimated above)
	  mod_comp = fft(abs(fpsi1_app).^2) ;
	  mod_comp = [mod_comp(1:round(nb/2)), ...
		      mod_comp(Nw-nb+round(nb/2)+1:Nw)] ;
	  mod_comp = sqrt(abs(ifft(mod_comp))) * sqrt(Nw/nb) ;
	  % difference between each line of fpsi1_from_sol2 and
	  % mod_comp
	  diff = bsxfun(@minus,abs(fpsi1_from_sol2),mod_comp) ;
	  norm_diff = sum(diff.^2,2) ;
	  [forg,ind_sort] = sort(norm_diff,'ascend') ;

	  % maximal number of candidates retained in sol2
	  lim2 = min(size(sol1,1),size(sol2,1)) ;
	  % keep the ones which best approximate mod_comp
	  sol2 = sol2(ind_sort(1:lim2),:) ;
	  % for any of these candidates, compute the corresponding
	  % (f*win) star psi2 star psi1 (in Fourier)
	  fpsi2_t = bsxfun(@times,sol2(:,2:size(sol1,2)), ...
			   psi1_f(2:size(sol1,2))) ;
	  % for each element of sol1, compute the corresponding
	  % (f*win) star psi1 star psi2 (in Fourier)
	  fpsi1_t = bsxfun(@times,sol1(:,2:end), ...
			   psi2_f(2:size(sol1,2))) ;
	  % distance between each line of fpsi2_t and fpsi2_t (up to
	  % multiplication by a unitary complex)
          scals = fpsi1_t * fpsi2_t' ;
          mat_dist = bsxfun(@plus,bsxfun ...
				    (@plus,-2*abs(scals), ...
				     sum(abs(fpsi1_t).^2,2)), ...
                            sum(abs(fpsi2_t).^2,2)') ;
	  % divide by the norm of fpsi1_t (to avoid giving a
	  % preference to small-norm solutions)
	  mat_dist = bsxfun(@rdivide,mat_dist, ...
			    sqrt(sum(abs(fpsi1_t).^2,2))) ;
	  % indexes where the distance is minimal
          [forg,ind_opt1] = min(min(mat_dist,[],2)) ;
	  [forg,ind_opt2] = min(mat_dist(ind_opt1,:)) ;
            
            
	  if debug.bool && strcmp(debug.function,'reestimate') ...
	     && (S == debug.num_params(1)) ...
	     && (s == debug.num_params(2)) ...
	     && (s2 == debug.num_params(3))	     
	    % debugging

	    fpsi1_t_tmp = fft(fpsi1_rec_tmp) ;
	    fpsi1_t_tmp = fpsi1_t_tmp(2:size(sol1,2)) .* ...
			  psi2_f(2:size(sol1,2)) ;
	    fpsi2_t_tmp = fft(fpsi2_rec_tmp) ;
	    fpsi2_t_tmp = fpsi2_t_tmp(2:size(sol1,2)) .* ...
			  psi1_f(2:size(sol1,2)) ;
	    fpsi_12_true = fft(f_trunc) ;
	    fpsi_12_true = fpsi_12_true(2:size(sol1,2)) .* ...
			   psi1_f(2:size(sol1,2)) .* ...
			   psi2_f(2:size(sol1,2)) ;
	    figure (3) ; clf ;
	    subplot(121) ; hold on ;
	    plot(abs(fpsi1_t(ind_opt1,:))) ;
	    plot(abs(fpsi2_t(ind_opt2,:)),'r') ;
	    plot(abs(fpsi_12_true),'g') ;
	    title(['f star psi1 star psi2 (optimal solutions) ; ', ...
		   'truth (green), sol1 (blue), sol2 (red)']) ;
	    subplot(122) ; hold on ;
	    plot(abs(fpsi1_t_tmp)) ;
	    plot(abs(fpsi2_t_tmp),'r') ;
	    plot(abs(fpsi_12_true),'g') ;
	    title(['f star psi1 star psi2 (solutions of fig 2) ; ', ...
		   'truth (green), sol1 (blue), sol2 (red)']) ;

            fpsi1_rec_tmp = ifft(sol1(ind_opt1,:),Nw) ;
            fpsi1_rec_tmp = fpsi1_rec_tmp * ...
			    exp(-i*angle(fpsi1_rec_tmp(Nw/2)));
            
	    figure (4) ; clf ;
            subplot(131) ; hold on ;
            plot(abs(fft(fpsi1_true))) ;
            plot(abs(fft(fpsi1_rec_tmp)),'r') ;
            subplot(132) ; hold on ;
            plot(real(fpsi1_true)) ;
            plot(real(fpsi1_rec_tmp),'r') ;
	    title(['true (blue) and reconstructed (red) ', ...
		   'fw star psi1 (Fourier, time, phase diff)']) ;
            subplot(133) ; hold on ;
            plot(unwrap(angle(fpsi1_rec_tmp ./ ...
                              fpsi1_true))) ;

          end
            
	  % optimal choice for (f*win) star psi1
          fpsi1_opt = ifft(sol1(ind_opt1,:),Nw) ;
	  % deduce f star psi1 (on the support of win)
	  fpsi1_opt = fpsi1_opt(max(1,min(Nw,[1:Nw]+ ...
					     round(shifts1_inv)))) ...
		      .* mul1_inv ;
	  % keep only the center (on the sides of the window, the
	  % error is too large)
          fpsi1_rec{s2} = fpsi1_opt(round(Nw/2)- ...
                                    round(2*spa/3):round(Nw/2)+...
						round(2*spa/3)) ;

        end

        fprintf('\n') ;

	% aggregate the small reestimations to obtain the reestimation
	% of the whole block
        subsize = 2*round(2*spa/3)+1 ; % size of each reestimated
	                               % segment
        I1(s) = inds_c(1)-round(2*spa/3) ; % beginning of the block
        I2(s) = inds_c(numel(inds_c))+round(2*spa/3) ; % end
        numel_inds(s) = numel(inds_c) ;
        % initialize the reestimated block
	fpsi1_block{s} = zeros(1,I2(s)-I1(s)+1) ;
	% copy in it the first reestimated segment
        fpsi1_block{s}(1:subsize) = fpsi1_rec{1} ;
	
	% for each reestimated segment, copy it in fpsi1_block{s} (and
	% take care of eventual discontinuities which may occur
	% between consecutive blocks)
        for s2=2:numel(inds_c)
            
          i1 = inds_c(s2)-round(2*spa/3) -I1(s) +1 ;
          i2 = inds_c(s2)+round(2*spa/3) -I1(s) +1 ;
          i2_prev = inds_c(s2-1)+round(2*spa/3) -I1(s) +1 ;
            
	    % subsegment of fpsi1_rec{s2} which intersects the
	    % previous segment
	    inter_fpsi = fpsi1_rec{s2}(1:i2_prev-i1+1) ;
            inter_fpsi_prev = fpsi1_rec{s2-1}(end-(i2_prev-i1):end) ;
            
	    % multiply fpsi1_rec{s2} by a global phase and update
	    % inter_fpsi
            alpha = angle(sum(inter_fpsi_prev.*conj(inter_fpsi))) ;
            fpsi1_rec{s2} = exp(i*alpha) * fpsi1_rec{s2} ;
            inter_fpsi = fpsi1_rec{s2}(1:i2_prev-i1+1) ;
            
	    % copy in fpsi1_block the part of fpsi1_rec{s2} which does
	    % not intersect the previous segment
            fpsi1_block{s}(i2_prev+1:i2) = fpsi1_rec{s2} ...
						    (i2_prev-i1+2:end) ;
            
	    % on the part of fpsi1_block where there is a conflict
	    % between the segment and the previous one, copy a
	    % weighted sum of the two segments
            [win1,win2] = double_window(i2_prev-i1+1) ;
            fpsi1_block{s}(i1:i2_prev) = win1 .* inter_fpsi_prev ...
					 + win2 .* inter_fpsi ;
            
        end
        
        if (I2(s)-I1(s)>=N)
	  % case where the block recovers all the signal

	  % the beginning and the end of fpsi1_block represent the
	  % same part of the signal
          fpsi_beg = fpsi1_block{s}(1:I2(s)-I1(s)-N+1) ;
          fpsi_end = fpsi1_block{s}(N+1:end) ;
            
	  % multiply the reestimated block by a very low-frequency
	  % sinusoid
          alpha = angle(sum(fpsi_beg.*conj(fpsi_end))) ;
          fpsi1_block{s} = fpsi1_block{s} .* ...
			   exp(i*alpha*[0:I2(s)-I1(s)]/N) ;
	  % update beginning and end
          fpsi_beg = fpsi1_block{s}(1:I2(s)-I1(s)-N+1) ;
          fpsi_end = fpsi1_block{s}(N+1:end) ;
            
	  % replace the beginning and the end by a linear combination
	  % of both
          [win1,win2] = double_window(I2(s)-I1(s)-N+1) ;
          fpsi1_block{s}(1:I2(s)-I1(s)-N+1) = ...
          win1 .* fpsi_end + win2 .* fpsi_beg ;
          fpsi1_block{s}(N+1:end) = ...
          win1 .* fpsi_end + win2 .* fpsi_beg ;
          
	  % deduce a reestimation of f star psi1
          fpsi1_rees = fpsi1_block{s}(mod([2-I1(s):2-I1(s)+N-1]-1, ...
					 I2(s)-I1(s)+1)+1) ;
          
	  % from f star psi1, deduce a reestimation of f star
	  % psi_low(S,:) and of f star psi_high(S+1,:)
          mult = psi_low(S,:) ./ ...
                 (psi(S,:) + (abs(psi(S,:))<1e-12)) ;
          fpsi_low_rees = ifft(fft(fpsi1_rees).*mult) ;
          mult = psi_high(S+1,:) ./ ...
                 (psi(S,:)+(abs(psi(S,:))<1e-12)) ;
          fpsi_high_rees = ifft(fft(fpsi1_rees).*mult) ;
          
          % multiply the reestimated f star psi_low and f star
	  % psi_high by low frequency phases : the reestimation
	  % provided by the previous code may be locally correct but
	  % it is not reliable up to multiplication by a low
	  % frequency.
          n = floor(numel(inds_c)/8) ;
          fpsi_low(S,:) = slow_phase_mult(fpsi_low_rees, ...
					  fpsi_low(S,:),n) ;
          fpsi_high(S+1,:) = slow_phase_mult(fpsi_high_rees, ...
					     fpsi_high(S+1,:),n) ;

          % all the signal has been reestimated ; exit.
          not_break = 0 ;
          break ;
        
	end
        
      end
    
      if not_break
	% In the previous part of this function, an estimation of f
	% star psi(S,:) has been computed in several points, grouped
	% in blocks.
	% Now, from this estimation, we deduce estimations for f star
	% psi_low(S,:) and f star psi_high(S,:).

	% from the current value of fpsi_low(S,:), compute an
	% approximation of fpsi_high(S,:) then an approximation of f
	% star psi(S,:)
        fpsi_high_rough = conj(products(S,:)./fpsi_low(S,:)) ;
        div = abs(psi_low(S,:)).^2 + ...
              abs(psi_high(S,:)).^2 ;
        fpsi_rough = (fft(fpsi_high_rough).* ...
		      conj(psi_high(S,:)) + ...
		      fft(fpsi_low(S,:)).*conj(psi_low(S,:))) .* ...
		     psi(S,:) ./ (div + (div<1e-12)) ;
        fpsi_rough = ifft(fpsi_rough) ;
        
        for s=1:nb_blocks
	  
	  % for each block, the reestimation is not correct up to
	  % multiplication by a low-frequency phase ; multiply each
	  % block by a low-frequency phase, to 'align' it with its
	  % non-reestimated version
  
	  % copy of fpsi_rough
          fpsi_rough_cp = fpsi_rough ;
	  
	  % copy the reestimated block in fpsi_rough_cp

	  % adjusts the phase (to avoid discontinuities at
	  % the boundary)
	  fpsi1_block{s} = ...
	  adjust_phase ...
	    (fpsi1_block{s}, ...
	     fpsi_rough_cp(mod([I1(s):inds_beg(s)]-1,N)+1), ...
	     fpsi_rough_cp(mod([inds_end(s):I2(s)]-1,N)+1)) ;
	  
	  % copy the beginning of the block
          [win1,win2] = double_window(inds_beg(s)-I1(s)+1) ;
          fpsi_rough_cp(mod([I1(s):inds_beg(s)]-1,N)+1) = ...
          win1 .* fpsi_rough_cp(mod([I1(s):inds_beg(s)]-1,N)+1) + ...
          win2 .* fpsi1_block{s}(1:inds_beg(s)-I1(s)+1) ;
	  % copy the middle
          fpsi_rough_cp(mod([inds_beg(s)+1:inds_end(s)-1]-1,N)+1) = ...
          fpsi1_block{s}(inds_beg(s)-I1(s)+2: ...
                         inds_end(s)-I1(s)) ;
	  % copy the end
          [win1,win2] = double_window(I2(s)-inds_end(s)+1) ;
          fpsi_rough_cp(mod([inds_end(s):I2(s)]-1,N)+1) = ...
          win1 .* fpsi1_block{s}(end-I2(s)+inds_end(s):end) + ...
          win2 .* fpsi_rough_cp(mod([inds_end(s):I2(s)]-1,N)+1) ;

	  % from fpsi_rough_cp, compute an estimation of ...
	  % f star psi_high(S+1,:)
          mult = psi_high(S+1,:) ./ ...
                 (psi(S,:)+(abs(psi(S,:))<1e-12)) ;
          fpsi_high_rees = ifft(fft(fpsi_rough_cp).*mult) ;
            
          % adjusts the phase
          n = floor(numel_inds(s)/8) ;
          inds_tmp = mod([I1(s):I2(s)]-1,N)+1 ;
	  [forg,phase] = slow_phase_mult(fpsi_high_rees(inds_tmp), ...
					 fpsi_high(S+1,inds_tmp),n,'not_int') ;
	  fpsi1_block{s} = phase .* fpsi1_block{s} ;
            
        end
        
	% copy the reestimated blocks in fpsi_rough
        for s=1:nb_blocks

	  % avoid boundary problems
          fpsi1_block{s} = ...
	  adjust_phase ...
	    (fpsi1_block{s}, ...
	     fpsi_rough(mod([I1(s):inds_beg(s)]-1,N)+1), ...
	     fpsi_rough(mod([inds_end(s):I2(s)]-1,N)+1)) ;

	  % copy the beginning of the block
	  d = inds_beg(s)-I1(s) ;
	  [win1,win2] = double_window(d) ;
	  fpsi_rough(mod([I1(s):inds_beg(s)-1]-1,N)+1) = ...
	  fpsi_rough(mod([I1(s):inds_beg(s)-1]-1,N)+1) .* ...
	  win1 + fpsi1_block{s}(1:d) .* win2 ;
	  % copy the middle
	  fpsi_rough(mod([inds_beg(s):inds_end(s)]-1,N)+1) = ...
	  fpsi1_block{s}(d+1:d+1+inds_end(s)-inds_beg(s)) ;
	  % copy the end
	  d = I2(s)-inds_end(s) ;
	  [win1,win2] = double_window(d) ;
	  fpsi_rough(mod([inds_end(s)+1:I2(s)]-1,N)+1) = ...
	  fpsi1_block{s}(end-d+1:end) .* win1 + ...
	  fpsi_rough(mod([inds_end(s)+1:I2(s)]-1,N)+1) .* win2 ;
          
        end
        
	% deduce reestimated versions of f star psi_low(S,:) and f
	% star psi_high(S+1,:)
        mult = psi_low(S,:) ./ ...
               (psi(S,:) + (abs(psi(S,:))<1e-12)) ;
        fpsi_low_rees = ifft(fft(fpsi_rough).*mult) ;
        mult = psi_high(S+1,:) ./ ...
               (psi(S,:) + (abs(psi(S,:))<1e-12)) ;
        fpsi_high_rees = ifft(fft(fpsi_rough).*mult) ;
        
	% replace the uncorrect values in fpsi_low(S,:) by their
	% reestimation
        for s=1:nb_blocks
            
	  i_inf = inds_beg(s)-spa ;
	  i_sup = inds_end(s)+spa ;

	  % avoid boundary problems
          fpsi_low_rees(mod([i_inf:i_sup]-1,N)+1) = ...
	  adjust_phase ...
            (fpsi_low_rees(mod([i_inf:i_sup]-1,N)+1), ...
	     fpsi_low(S,mod([i_inf:inds_beg(s)-1]-1,N)+1), ...
	     fpsi_low(S,mod([inds_end(s)+1:i_sup]-1,N)+1)) ;
            
	  % copy the middle
          fpsi_low(S,mod([inds_beg(s):inds_end(s)]-1,N)+1) = ...
          fpsi_low_rees(mod([inds_beg(s):inds_end(s)]-1,N)+1) ;            
	  % copy the beginning
          [win1,win2] = double_window(inds_beg(s)-i_inf) ;
          fpsi_low(S,mod([i_inf:inds_beg(s)-1]-1,N)+1) = ...
          win1 .* fpsi_low(S,mod([i_inf: ...
                                  inds_beg(s)-1]-1,N)+1) + ...
          win2 .* fpsi_low_rees(mod([i_inf: ...
                                     inds_beg(s)-1]-1,N)+1) ;
          % copy the end
          [win1,win2] = double_window(i_sup-inds_end(s)) ;
          fpsi_low(S,mod([inds_end(s)+1:i_sup]-1,N)+1) = ...
          win1 .* fpsi_low_rees(mod([inds_end(s)+1: ...
                                     i_sup]-1,N)+1) + ...
          win2 .* fpsi_low(S,mod([inds_end(s)+1:i_sup]-1,N)+1) ;
        end

	% same thing in fpsi_high(S+1,:)
        for s=1:nb_blocks

	  i_inf = inds_beg(s) - spa ;
          i_sup = inds_end(s) + spa ;

	  % avoid boundary problems
          fpsi_high_rees(mod([i_inf:i_sup]-1,N)+1) = ...
	  adjust_phase ...
            (fpsi_high_rees(mod([i_inf:i_sup]-1,N)+1), ...
	     fpsi_high(S,mod([i_inf:inds_beg(s)-1]-1,N)+1), ...
	     fpsi_high(S,mod([inds_end(s)+1:i_sup]-1,N)+1)) ;
          
	  % copy the middle of the block
          fpsi_high(S+1,mod([inds_beg(s):inds_end(s)]-1,N)+1) = ...
          fpsi_high_rees(mod([inds_beg(s):inds_end(s)]-1,N)+1) ;
          % copy the beginning
          [win1,win2] = double_window(inds_beg(s)-i_inf) ;
          fpsi_high(S+1,mod([i_inf:inds_beg(s)-1]-1,N)+1) = ...
          win1 .* fpsi_high(S+1,mod([i_inf: ...
                                     inds_beg(s)-1]-1,N)+1) + ...
          win2 .* fpsi_high_rees(mod([i_inf: ...
                                      inds_beg(s)-1]-1,N)+1) ;
          % copy the end
          [win1,win2] = double_window(i_sup-inds_end(s)) ;
          fpsi_high(S+1,mod([inds_end(s)+1:i_sup]-1,N)+1) = ...
          win1 .* fpsi_high_rees(mod([inds_end(s)+1: ...
                                      i_sup]-1,N)+1) + ...
          win2 .* fpsi_high(S+1,mod([inds_end(s)+1:i_sup]-1,N)+1) ;
        end

      end
    
      if debug.bool && strcmp(debug.function,'reestimate') ...
	 && (S == debug.num_params(1))
	% debugging
        
	fprintf('Beginning of blocks:') ;
        inds_beg
	fprintf('End of blocks:') ;
        inds_end
        
        figure (5) ; clf ;
	subplot(121) ; hold on ;
	fpsi_low_true = wavelet_transform(debug.true_f,psi_low(S,:)) ;
        plot(unwrap(angle(fpsi_low(S,:)./fpsi_low_true))) ;
        plot(unwrap(angle(fpsi_low_init./fpsi_low_true)),'g') ;
	title(['Phase difference between true and initial (green) ', ...
	       '/ reestimated (blue) fpsi low']) ;
	subplot(122) ; hold on ;
        plot(abs(fpsi_low_true)) ;
	plot(abs(fpsi_low(S,:)),'r') ;
	plot(abs(fpsi_low_init),'g')
	title(['True (blue), initial (green) and reestimated (red)', ...
	       ' fpsi low in modulus']) ;
 
        error('reestimate.m: stopped for debugging') ;

    end
    
end
end
