function sol = solutions(modu,psi,nlim) ;
  % Computes all the possible solutions to the following equation :
  %             abs(ifft(sol)) = modu
  % such that the support of sol is included in the support of psi

  % The support is assumed to be of the forme {2,...,i}.
  % Let P be the polynom P(X) = sum_(2<=j<=i) sol(j) X^(j-2)
  % If we denote by Q the conjugate of P, then, from abs(ifft(sol)),
  % we can compute P(X)Q(1/X). From this, we can deduce the possible
  % P's. For more details, see for example Sanz, 'Mathematical
  % considerations for the problem of fourier transform phase
  % retrieval from magnitude'

  % nlim is an integer such that the number of proposed solutions must
  % not exceed 2^nlim

  N = size(psi,2) ;
  
  % size of the support of psi
  max_psi = max(abs(psi)) ;
  ind2 = max([1:N/2].*(abs(psi(1:floor(N/2)))>0.02*max_psi)) ;

  % the solutions returned will have a support of size at most ind2
  % in pratice, it seems that a smaller size may be beneficial if
  % possible; for all sizes ind smaller than ind2, check whether it is
  % possible to find signals with a support of size ind, whose inverse
  % Fourier transform has a modulus equal to modu
  modu_f = fft(abs(modu).^2) ;
  for ind = 1:ind2
    modu_f_test = modu_f ;
    modu_f_test(ind:min(N,N-ind+2)) = 0 ;
    modu_test = sqrt(max(0,real(ifft(modu_f_test)))) ;
    % modu_test is (approximately) the modulus of the inverse Fourier
    % transform of the computed solutions if these solutions are
    % constrained to have a support of size ind
    if (max(abs(abs(modu)-modu_test)) < 0.02 * max(abs(modu)))
      % if modu_test is close to modu, then keep the value ind
      break ;
    end
  end

  % if ind is too large, the number of roots will exceed nlim and
  % there will be too many solutions
  ind = min(ind,nlim+2) ;
  nb_shift = min(ind2)-ind ; % number of different possible positions
                             % for the support of the solutions
  % the number of solutions is bounded by (nb_shift+1)*2^ind
  % we want this number to be at most 2^nlim
  while ((nb_shift+1)*2^ind)>2^nlim
    ind = ind-1 ;
    nb_shift = nb_shift+1 ;
  end

  % outside the support, the coefficients are assumed to be noise  
  modu = real(ifft(modu_f)) ;
  modu = modu - min(0,min(modu)) ;
  modu_f = fft(modu) ;
  modu_f(ind:min(N,N-ind+2)) = 0 ;

  % coefficients of P(X)Q(1/X)
  coeffs = [modu_f(min(N,N-ind+2)+1:N), ...
            modu_f(1:ind-1)]*N ;
  % roots of this polynom
  root = roots(fliplr(coeffs)) ;
    
  % roots with modulus different from 1
  root_t = root(find(abs(abs(root)-1)>1e-4)) ;
  % roots with modulus 1 (if there was no noise, all these roots
  % would have an even multiplicity)
  root_w = root(find(abs(abs(root)-1)<=1e-4)) ;
  % roots with modulus larger than one
  root = root_t(find(abs(root_t)>1)) ;
  ind2 = numel(root) ;

  % replaces root_w by a set with two times less elements such that
  % the initial root_w is "as much as possible equal" to the
  % concatenation of the final set with itself
  if (numel(root_w)>0)
    [forg,ind_w] = sort(angle(root_w)) ;
    root_w1 = root_w(ind_w(1:2:end)) ;
    root_w2 = root_w(ind_w(2:2:end)) ;
    if (sum(abs(angle(root_w1./root_w2))) > ...
        sum(abs(angle(root_w1./[root_w2(end);root_w2(1:end-1)]))))
      root_w = root_w1 .* sqrt([root_w2(end); ...
					    root_w2(1:end-1)]./root_w1) ;
    else
      root_w = root_w1 .* sqrt(root_w2./root_w1) ;
    end
  end

  % the solutions are the polynomial functions whose roots are the
  % elements of root and root_w but each element of root can
  % eventually be replaced by its conjugate inverse. This gives 2^ind2
  % solutions, where ind2 is the number of elements in root.
  st = dec2bin([0:2^ind2-1],ind2) ;
  inds = find (st == '1') ;
  root_s = repmat(root.',[2^ind2,1]) ;
  root_s(inds) = 1./conj(root_s(inds)) ;
  % leading coefficient
  if ~isempty(root_s)
    prod_root_s = prod(abs(root_s),2) ;
  else
    prod_root_s = [1] ;
  end
  alphas = sqrt(abs(coeffs(end))./prod_root_s) ...
           / sqrt(prod(abs(root_w))) ;
  root_tot = [root_s,repmat(root_w.',[2^ind2,1])] ;
  sol = [zeros(2^ind2,1), bsxfun(@times,poly_mat(root_tot), ...
				 alphas)] ;

  if (nb_shift>0)
    % tries nb_shift different positions for the support of the
    % solutions
    sol_tmp = [] ;
    for k=0:nb_shift
      sol_tmp = [sol_tmp ;
		 [zeros(size(sol,1),k), sol, ...
		  zeros(size(sol,1),nb_shift-k)]] ;
    end
    sol = sol_tmp ;
  end
