%% initialize the random generator
seed = 116025 ;
rand('state',seed) ;
randn('state',seed) ;

%% Parameters
N = 500 ; % size of the signal
signal_type = 'piecewise_regular' ;
wavelet_type = 'simili-morlet' ; % name of the wavelet family
noise_type = 'simple_gaussian' ; % type of noise
a = 2 ; % dilation factor for the wavelet family
ln = 1e-3 ; % relative norm of the noise

%% choose a signal
f = create_signal(N,signal_type) ;
% for an audio signal, uncomment the following two lines
% [f,freq_sample] = create_signal([],'audio','source_audio/commons_sorry.wav') ;
% N = size(f(:),1) ;

%% generate wavelets
[freqs_wav,psi] = create_wavelets(N,a,wavelet_type) ;
nbf = size(freqs_wav,2) ; % number of wavelets in the family

%% wavelet transform
fpsi = wavelet_transform(f,psi) ;
% add noise
fpsi_noisy = add_global_noise(ln,noise_type,abs(fpsi),[],psi) ;

%% reconstruction
% options for the reconstruction algorithm
opts.time_limit = Inf ; % maximal time spent for each gradient descent
                       % (in seconds)
opts.Kit_descent = 10000 ; % maximal number of iterations
opts.Kit_gs_rough = 10 ; % number of Gerchberg-Saxton iterations for the
                         % first rough approximation
opts.Kit_gs = 2000 ; % number of Gerchberg-Saxton iterations for the
                     % final refinement
opts.nb_print_descent = 1000 ; % rate of display
                               % during the optimization step
opts.nb_print_gs = 500 ; % rate of display during Gerchberg-Saxton
opts.reestimate = true ; % follow the reestimation procedure
opts.only_gs = false ; % if true, reconstruct by Gerchberg-Saxton
                       % (for the purposes of comparison)
% informations about the noise
noise_descr.ln = ln ;
noise_descr.noise_type = noise_type ;

% debugging parameters
debug.bool = false ; % whether to perform debugging or not
debug.function = 'reconstruct' ; % which function must be debugged
debug.num_params = [11,3] ; % parameters for the debugging
debug.true_f = f ; % true signal

% reconstruction
f_rec = reconstruct(fpsi_noisy,psi,freqs_wav,a, ...
		    noise_descr,opts,debug) ;

%% compute the errors
% multiplication by a global phase
al = exp(i*angle(sum(f(:).*conj(f_rec(:))))) ;
f_rec = al*f_rec ;
% wavelet transform of the reconstructed function
fpsi_rec = wavelet_transform(f_rec,psi) ;
diff_abs = abs(fpsi)-abs(fpsi_rec) ;
fprintf('Relative error on the moduli : %.2e\n', ...
        norm(diff_abs(:),2)/norm(fpsi(:),2)) ;
fprintf('Relative error on the function : %.2e\n', ...
        norm(f(:)-f_rec(:),2)/norm(f,2)) ;

%% display the results
figure (1) ; clf ; hold on ;
plot(real(f),'b') ;
plot(real(f_rec),'r') ;
title('Initial and reconstructed functions')

figure (2) ; clf ; hold on ;
draw_functions(abs(fpsi_noisy),'g') ;
draw_functions(abs(fpsi),'b') ;
draw_functions(abs(fpsi_rec),'r') ;
ax = findobj(gcf,'type','axes');
pos = cell2mat(get(ax,'position'));
ncols = numel(unique(pos(:,1)));
nrows = numel(unique(pos(:,2)));
subplot(nrows,ncols,ceil(ncols/2)) ;
title('Initial (blue), noisy (green) and reconstructed (red) moduli');

figure (3) ; clf ;
image_angle(fpsi_rec,fpsi,1e-2) ;
title('Difference in phase between the initial and reconstructed functions')

%% save audio signal to file
% out_name = 'rec_sorry.wav' ;
% f_rec = 2 * real(f_rec) ;
% if (max(abs(f_rec))>1)
%   f_rec = f_rec / max(abs(f_rec)) ; % to avoid clipping
% end
% audiowrite(out_name,f_rec,freq_sample) ;
